structure LambdaPlusPlus =
struct
  datatype typ =
    BaseType of string
  | Arrow of typ * typ
  | Plus of typ * typ
  | Times of typ * typ

  datatype exp =
    Variable of string
  | Lambda of (string * typ) * exp
  | Apply of exp * exp
  | Tuple of exp * exp
  | First of exp
  | Second of exp
  | Left of exp * typ
  | Right of exp * typ
  | Case of exp * (string * exp) * (string * exp)

  infix 6 ^^
  fun s1 ^^ s2 = s1 ^ " " ^ s2
  val rec pp_typ = (fn
      BaseType a => a
    | Arrow (t1, t2) => "(" ^ pp_typ t1 ^^ "->" ^^ pp_typ t2 ^ ")"
    | Plus (t1, t2) => "(" ^ pp_typ t1 ^^ "+" ^^ pp_typ t2 ^ ")"
    | Times (t1, t2) => "(" ^ pp_typ t1 ^^ "*" ^^ pp_typ t2 ^ ")"
  )

  val rec pp_exp = (fn
      Variable x => x
    | Lambda ((x, t), e) => "fn (" ^ x ^^ ":" ^^ pp_typ t ^ ") =>" ^^ pp_exp e
    | Apply (e1, e2) => "(" ^ pp_exp e1 ^^ pp_exp e2 ^ ")"
    | Tuple (e1, e2) => "(" ^ pp_exp e1 ^ "," ^^ pp_exp e2 ^ ")"
    | First e => "#1 (" ^ pp_exp e ^ ")"
    | Second e => "#2 (" ^ pp_exp e ^ ")"
    | Left (e, t) => "INL" ^^ pp_exp e ^^ "into _ +" ^^ pp_typ t
    | Right (e, t) => "INR" ^^ pp_exp e ^^ "into" ^^ pp_typ t ^^ "+ _"
    | Case (e, (x1, e1), (x2, e2)) =>
      "case" ^^ pp_exp e ^^ "of" ^^
      "INL" ^^ x1 ^^ "=>" ^^ pp_exp e1 ^^ "|" ^^
      "INR" ^^ x2 ^^ "=>" ^^ pp_exp e2
  )
end
